// ============================================================================
//
// Copyright (c) 2010-2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ============================================================================


#include "Board.h"

#include <cstdlib>
#include <list>
#include <sstream>
#include <string>
#include <vector>

#include "Cell.h"
#include "Common.h"
#include "Generator.h"
#include "Grid.h"
#include "MainFrame.h"
#include "SolverThread.h"


BEGIN_EVENT_TABLE(Board, wxPanel)
    EVT_SIZE(Board::OnSize)
    EVT_PAINT(Board::OnPaint)
    EVT_LEFT_DOWN(Board::OnLeftDown)
END_EVENT_TABLE();


Board::Board(MainFrame *parent)
      :wxPanel(parent, wxID_ANY)
{
    this->parent = parent;
    grid = NULL;
    gridForDisplay = NULL;
    solver = NULL;
    generator = NULL;
    log = NULL;

    size = -1;

    wxWindow::SetBackgroundStyle (wxBG_STYLE_CUSTOM);
	
    this->normalFont = wxFont::New (16, (wxFontFamily)wxFONTFAMILY_MODERN,
             wxFONTSTYLE_NORMAL, (wxFontWeight)wxFONTWEIGHT_BOLD,
             false, std2wx(""), wxFONTENCODING_DEFAULT);

    this->smallFont = wxFont::New (9, (wxFontFamily)wxFONTFAMILY_MODERN,
             wxFONTSTYLE_NORMAL, (wxFontWeight)wxFONTWEIGHT_BOLD,
             false, std2wx(""), wxFONTENCODING_DEFAULT);	

    SetAutoLayout (true);
}

void Board::resolvingFinished()
{
	wxCommandEvent event( wxEVT_COMMAND_TEXT_UPDATED, 100002 );
	parent->GetEventHandler()->AddPendingEvent( event );
}


void Board::initGrid()
{
    // Might want to change the board size, hence decomment
    //size = getSizeFromUser();
    size = 9;

    // Convert int to string
    std::string intAsString;
    std::stringstream convStream;
    convStream << size;
    convStream >> intAsString;

    std::string sizeStr = "Size : ";
    std::string concat = sizeStr + intAsString;

    wxString textAsString = std2wx(concat);

    parent->setStatusText (textAsString);

    this->generator = new Generator (this, size);
    this->grid = new Grid (size);
}

MainFrame* Board::getMainFrame()
{
    return parent;
}

void Board::setGrid (Grid* grid)
{
    this->grid = grid;
}

void Board::generateGrid (int difficulty)
{
    Grid* newGrid = generator->generateGrid(difficulty);
    if (this->grid != NULL)
    {
       delete this->grid;
    }
    this->grid = newGrid;

	this->Refresh();
}

void Board::cleanGrid ()
{
    if (this->grid != NULL)
    {
       delete this->grid;
    }

    this->grid = new Grid (size);
    this->Refresh();
}

void Board::resetGrid ()
{
    if (this->grid != NULL)
    {
       this->grid->reset();
    }

    this->Refresh();
}


/*
 * Intended to be used by non-GUI threads only !!!
 */
void Board::setStatusText (const wxString& text)
{
	wxCommandEvent event( wxEVT_COMMAND_TEXT_UPDATED, 100001 );
	event.SetString(text);
	parent->GetEventHandler()->AddPendingEvent( event );
}

int Board::getSize()
{
    return size;
}

int Board::getSizeFromUser()
{
    int retValue = 0;
    do
    {
        const char* input = getInputFromUser ("Enter Grid size", "Grid size input", "9", false).mb_str();
        retValue = atoi (input);

    } while (retValue == 0);

    return retValue;
}

int Board::getValueFromUser()
{
    const char* input = getInputFromUser (
            "Enter Value for Cell\n - Enter Single value or \n - Several possibilities separated by space",
            "Cell value input", "", true).mb_str();
    return atoi (input);
}

void Board::showError (const char* message)
{
    /*
    wxMessageDialog errorDlg(
            this, std2wx (message), std2wx ("Error"));
    errorDlg.Center();
    errorDlg.ShowModal();
    */
    this->setStatusText(std2wx (message));
}

wxString Board::getInputFromUser(const char* label, const char* title, const char* defaultValue, bool zeroAllowed)
{
    wxTextEntryDialog getSizeDialog (
            this, std2wx (label), std2wx(title), std2wx(defaultValue));
    getSizeDialog.SetValue(std2wx(defaultValue));
    getSizeDialog.Center();
    getSizeDialog.ShowModal();

    return getSizeDialog.GetValue();
}

void Board::OnSize(wxSizeEvent& event)
{
    this->Refresh();
}

void Board::redraw (wxSize size)
{
    this->SetSize(size);
    this->Refresh();
}

void Board::OnPaint(wxPaintEvent& event)
{
    wxAutoBufferedPaintDC dc (this);

    // clear the background
    dc.SetBackground(*wxGREY_BRUSH);
    dc.Clear();

    if (gridForDisplay != NULL)
    {
        gridForDisplay->draw (this, &dc, GetClientSize());
    }
    else
    {
        grid->draw (this, &dc, GetClientSize());
    }
}

void Board::initialize (InitValuesHolder* holder)
{
    grid->initialize (holder);
    this->Refresh();
}

InitValuesHolder* Board::getGridValuesHolder ()
{
    return grid->getGridValuesHolder ();
}

void Board::OnLeftDown(wxMouseEvent& event)
{
    wxPoint position = event.GetPosition();

    Cell* cell = grid->getCell (position);

    if (cell != NULL)
    {
        cell->setHighlight (true);
        this->Refresh();

        bool cont = true;
        while (cont)
        {
            try
            {
                // building initial value
                std::stringstream sstm;
                if (cell->hasValue())
                {
                    sstm << cell->getValue();
                }
                else
                {
                    std::vector<int>* poss = cell->getPossibilities ();
                    for (int i = 0; i < (int) poss->size(); i++)
                    {
                        sstm << poss->at (i) << ' ';
                    }
                }
                const char * c = sstm.str().c_str();

                //int value = this->getValueFromUser();
                wxString input = getInputFromUser (
                        "Enter Value for Cell\n - Enter Single value or \n - Several possibilities separated by space",
                        "Cell value input", c, true);

                cell->setValue (0);

                if (input.Find (' ') != wxNOT_FOUND)
                {
                    int index = -1;
                    int prevIndex = 0;
                    wxString remainder = input;
                    while ((index = remainder.Find (' ')) != wxNOT_FOUND)
                    {
                        wxString sub = remainder.SubString (0, index - 1);
                        int value = atoi (sub.mb_str());
                        if (value >= 0 && value <= 9)
                        {
                            cell->addPossibility(value);
                        }
                        prevIndex = index + 1;
                        remainder = remainder.SubString (prevIndex, remainder.Len() - 1);
                    }

                    wxString sub = remainder.SubString (0, index - 1);
                    if (sub.Len() > 0)
                    {
                        cell->addPossibility(atoi (sub.mb_str()));
                    }
                }

                // Test input : if several values separated by
                else
                {
                    //wxLogMessage ("Board::OnLeftDown  setting value %d", value);
                    int value = atoi (input.mb_str());
                    if (value >= 0 && value <= 9)
                    {
                        cell->setValue (value);
                    }
                }
                cont = false;
            }
            catch (SoldokuException* e)
            {
#ifdef __WXDEBUG__
                std::cout << e->what() << std::endl;
#endif
                delete e;
            }

        }

        cell->setHighlight (false);
        this->Refresh();
    }
}

void Board::start()
{
    stop();

    if (grid->isSolveableBacktrack())
    {
        if (solver != NULL)
        {
            delete solver;
            solver = NULL;
        }

        solver = new SolverThread(grid, this);
        solver->startResolving();
    }
    else
    {
    	wxCommandEvent event( wxEVT_COMMAND_TEXT_UPDATED, 100002 );
    	parent->GetEventHandler()->AddPendingEvent( event );
        this->showError("Grid is not solvable.");
    }
}

void Board::stop()
{
    if (solver != NULL)
    {
        solver->stopResolving();
    }
    delete solver;
    solver = NULL;
}

void Board::solve()
{
    if (grid->isSolveableBacktrack())
    {
        grid->start(new Solver (true));
    }
    else
    {
        this->showError("Grid is not solvable.");
    }
}

void Board::solveBacktrack()
{
    if (grid->isSolveableBacktrack())
    {
        Solver* solver = new Solver(true, NULL);

        std::list<Grid*>* solutions = solver->findSolutions(this->grid);
        if (solutions->size() == 0)
        {
            this->showError("No solutions");
        }
        else if (solutions->size() > 1)
        {
            this->showError("Many solutions !");
        }
        else
        {
            Grid* solvedGrid = solutions->front();
            for (int rowCounter = 0; rowCounter < size; rowCounter++)
            {
                for (int colCounter = 0; colCounter < size; colCounter++)
                {
                    Cell* cell = this->grid->getCell(rowCounter, colCounter);
                    Cell* sCell = solvedGrid->getCell(rowCounter, colCounter);
                    if (!cell->hasValue())
                    {
                        cell->setValue(sCell->getValue(), sdST_COMPUTED);
                    }
                }
            }

            solver->doRefresh();
        }

        while (solutions->size() > 0)
        {
            delete solutions->front();
            solutions->pop_front();
        }
        delete solutions;

        delete solver;
    }
    else
    {
        this->showError("Grid is not solvable.");
    }
}

void Board::OnKeyDown(wxKeyEvent& event)
{
    // TODO
}

void Board::OnTimer(wxCommandEvent& event)
{
    // TODO
}

Board::~Board()
{
    if (this->grid != NULL)
    {
        delete this->grid;
    }

    delete this->generator;
}

void Board::displayGrid (Grid* grid)
{
    if (grid == this->grid)
    {
        this->gridForDisplay = NULL;
    }
    else
    {
        this->gridForDisplay = grid;
    }
}

void Board::sleep ()
{
    usleep (20000);
}

void Board::doRefresh ()
{
	wxCommandEvent event( wxEVT_COMMAND_TEXT_UPDATED, 100000 );
	parent->GetEventHandler()->AddPendingEvent( event );
}

