// ============================================================================
//
// Copyright (c) 2010-2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ============================================================================


#include "Cell.h"

#include "Common.h"
#include "Solver.h"
#include "Board.h"



Cell::Cell(int rowIdx, int colIdx)
{
    this->value = 0;
    this->guess = 0;

    hasPossibilities = false;
    hasPreviousGuesses = false;

    this->color = *wxWHITE_BRUSH;
    this->cellType = sdST_INPUT;

    this->rowIdx = rowIdx;
    this->colIdx = colIdx;

    row = NULL;
    column = NULL;
    square = NULL;

    posX = -1;
    posY = -1;
    sizeX = -1;
    sizeY = -1;
}

Cell::~Cell()
{
}

const char* Cell::getName()
{
    std::string nameStr;
    std::ostringstream oss;
    oss << "Cell (" << (rowIdx + 1) << ", " <<  (colIdx + 1) << ")";
    nameStr = oss.str();

    return nameStr.c_str();
}

bool Cell::handleCellSinglePossibility(Solver *solver)
{
    bool singlePossiblity = false;

    if (!hasValue())
    {
        if (solver->shouldCont())
        {
            if (!solver->isFastResolution())
            {
                char statusStr[50];
                sprintf (statusStr, "Setting single possibility on Cell (%d, %d)", getRowIdx(), getColIdx());
                solver->setStatusText(std2wx(statusStr));
                highlight(solver);
            }

            // Loop through each number
            for (int current = 1; current <= row->getSize(); current++)
            {
                if (!getRow()->containsValue (current))
                {
                    if (containsValue (current))
                    {
                        //wxLogMessage ("Grid::solveSinglePossibilities - Cell (%d, %d) : ... already contains", rowCounter, colCounter);
                        singlePossiblity = true;
                    }
                    else
                    {
                        singlePossiblity = isSinglePossibility (current, row->getSize());
                        if (singlePossiblity)
                        {
                            //wxLogMessage ("Grid::solveSinglePossibilities - Cell (%d, %d) : SETTING VALUE !", rowCounter, colCounter);
                            setValue(current, sdST_COMPUTED);
                        }
                    }

                    if (singlePossiblity)
                    {
                        break;
                    }
                }
            }
        }
    }

    return singlePossiblity;
}


bool Cell::isSinglePossibility(int value, int size)
{
    if (containsValue (value))
    {
        return true;
    }

    // if pssibilities are set and value is the only candidate, return true
    int nArraySize = (int)possibilities.size();
    if (nArraySize == 1)
    {
        int singlePossibilityValue = possibilities[0];
        if (singlePossibilityValue == value)
        {
            return true;
        }
    }

    // First check amongst rows, columns and cells and see if single possibility
    if (isPossible (value))
    {
        if (row->isSinglePossibility(this, value))
        {
            return true;
        }
        else if (column->isSinglePossibility(this, value))
        {
            return true;
        }
        else if (square->isSinglePossibility(this, value))
        {
            return true;
        }
    }

    // Then proceed by exclusion

    // loop on each value
    // check if possible
    // if any OTHER value is possible, then single possibility is false
    bool retValue = true;
    for (int candidate = 1; candidate <= size; candidate++)
    {
        if (isPossible (candidate) && candidate != value)
        {
            retValue = false;
            break;
        }
    }

    return retValue;
}

int Cell::makeGuess()
{
    // 0. if any current guess add to already guesses
    if (guess != 0)
    {
        hasPreviousGuesses = true;
        previousGuesses.push_back(guess);
        this->freeValue(guess);
        guess = 0;
    }

    // 1. take first possibility
    if (possibilities.size() > 0)
    {
        // it is well possible in case of recursive guessing that this guess is
        // not possible any more since the value would have been taken by a
        // previous guess on same row / col / square => hence check it
        // (since guessing do not trigger a recomputing of possibilities)
        while (possibilities.size() > 0)
        {
            int candidate = possibilities.back();
            if (isPossible(candidate))
            {
                guess = candidate;
                break;
            }
            else
            {
                // discard possibility anyway
                discardPossibility (candidate);
            }
        }

        // if none were found, throw an exception
        if (guess == 0 && possibilities.size() <= 0)
        {
            throw new SoldokuException ((char*)"Unable to make any guess");
        }
    }

    // 2. Remove from possibilities
    if (guess != 0)
    {
        discardPossibility (guess);
        this->bookValue(guess);
    }

    return guess;
}

void Cell::fixGuess()
{
    setValue (getGuess(), sdST_FIXED);
}

bool Cell::allPossibilitiesTried()
{
    return possibilities.size() <= 0;
}

void Cell::addPossibility (int value)
{
    if (isPossible(value))
    {
        hasPossibilities = true;
        possibilities.push_back (value);
    }
}

bool Cell::setPossibilities (int size)
{
    hasPreviousGuesses = false;
    previousGuesses.clear();

    for (int value = 1; value <= size; value++)
    {
        if (isPossible(value))
        {
            hasPossibilities = true;
            possibilities.push_back (value);
        }
    }

    return false;
}

void Cell::resetPossibilities()
{
    hasPossibilities = false;
    hasPreviousGuesses = false;

    possibilities.clear();
    previousGuesses.clear();
}

bool Cell::discardPossibility(int candidateValue)
{
    // find the one to erase
    bool foundInPossibilities = false;
    if (hasPossibilities)
    {
        int possSize = possibilities.size();
        for (int i = 0; i < possSize; i++)
        {
            if (possibilities[i] == candidateValue)
            {
                //wxLogMessage ("Cell::discardPossibility - Cell (%d, %d) discarding possibility %d", rowNumber, colNumber, value);
                possibilities.erase (possibilities.begin() + i);
                foundInPossibilities = true;
                break;
            }
        }
        if (possibilities.size() <= 0)
        {
            hasPossibilities = false;
        }
    }

    if (!foundInPossibilities)
    {
        if (hasPreviousGuesses)
        {
            int guessesSize = previousGuesses.size();
            for (int i = 0; i < guessesSize; i++)
            {
                if (previousGuesses[i] == candidateValue)
                {
                    //wxLogMessage ("Cell::discardPossibility - Cell (%d, %d) discarding guess %d", rowNumber, colNumber, value);
                    previousGuesses.erase (previousGuesses.begin()+i);
                    foundInPossibilities = true;
                    break;
                }
            }
            if (previousGuesses.size() <= 0)
            {
                hasPreviousGuesses = false;
            }
        }
    }

    if (!foundInPossibilities)
    {
        if (guess == candidateValue)
        {
            this->freeValue(guess);
            guess = 0;
            foundInPossibilities = true;
        }
    }

    return foundInPossibilities;
}

bool Cell::hasPossibility (int possibility)
{
    if (hasPossibilities)
    {
        int possSize = possibilities.size();
        for (int i = 0; i < possSize; i++)
        {
            if (possibilities[i] == possibility)
            {
                return true;
            }
        }
    }

    if (hasPreviousGuesses)
    {
        int guessesSize = previousGuesses.size();
        for (int i = 0; i < guessesSize; i++)
        {
            if (previousGuesses[i] == possibility)
            {
                return true;
            }
        }
    }

    if (guess == possibility)
    {
        return true;
    }

    return false;
}

CellType Cell::getCellType ()
{
    return this->cellType;
}

void Cell::reset()
{
	if (this->cellType != sdST_INPUT)
	{
		setValue(0);
	}
	resetPossibilities();
}

void Cell::setRow (Row *row)
{
    if (rowIdx != row->getNumber())
    {
#ifdef __WXDEBUG__
        wxLogMessage (std2wx("throwing SoldokuException : Cell::setRow : provided row number is different than rowNumber"));
#endif
        throw new SoldokuException ((char*)"Cell::setRow : provided row number is different than rowNumber");
    }
    this->row = row;
}

void Cell::setColumn (Column *column)
{
    this->column = column;
}

void Cell::setSquare (Square *square)
{
    this->square = square;
}

Row* Cell::getRow ()
{
    return this->row;
}

Column* Cell::getColumn ()
{
    return this->column;
}

Square* Cell::getSquare ()
{
    return this->square;
}

void Cell::initPosition (wxCoord posX, wxCoord posY, wxCoord sizeX, wxCoord sizeY)
{
    this->posX = posX;
    this->posY = posY;
    this->sizeX = sizeX;
    this->sizeY = sizeY;
}

void Cell::draw (Board* board, wxDC* dc)
{
    dc->SetPen(*wxBLACK_PEN);
    dc->SetBrush(this->color);

    dc->DrawRectangle(posX, posY, sizeX, sizeY);

    int possSize = (int)possibilities.size();
    if (value != 0 || guess != 0)
    {
        //wxLogMessage ("Cell::draw  (%d, %d) -> %d", rowNumber, rowNumber, value);

        wxString str;
        if (value != 0)
        {
            switch (cellType)
            {
            case sdST_INPUT :
                dc->SetTextForeground(*wxBLACK);
                break;
            case sdST_COMPUTED :
                dc->SetTextForeground(*wxBLUE);
                break;
            case sdST_GUESS :
                dc->SetTextForeground(*wxRED);
                break;
            case sdST_POSS :
                dc->SetTextForeground(*wxGREEN);
                break;
            case sdST_FIXED :
                dc->SetTextForeground(*wxCYAN);
                break;
            }
            str.Printf(std2wx("%d"), value);
        }
        else
        {
            dc->SetTextForeground(*wxRED);
            str.Printf(std2wx("%d"), guess);
        }

        dc->SetFont (*(board->getNormalFont()));

        dc->DrawText(str,
                 (int) posX + (sizeX / 2) - 7,
                 (int) posY + (sizeY / 2) - 13);

    }
    else if (possSize > 0)// && possSize <= 4)
    {
        for (int possCount = 0; possCount < possSize; possCount++)
        {
            int possibility = possibilities[possCount];

            int possPosX = 0;
            int possPosY = 0;

            int sizeFontX = 4;
            int sizeFontY = 5;

            switch (possibility)
            {
            case 1:
                possPosX = posX + (sizeX / 4) - sizeFontX;
                possPosY = posY + (sizeY / 4) - sizeFontY;
                break;
            case 2:
                possPosX = posX + (2 * sizeX / 4) - sizeFontX;
                possPosY = posY + (sizeY / 4) - sizeFontY;
                break;
            case 3:
                possPosX = posX + (3 * sizeX / 4) - sizeFontX;
                possPosY = posY + (sizeY / 4) - sizeFontY;
                break;

            case 4:
                possPosX = posX + (sizeX / 4) - sizeFontX;
                possPosY = posY + (2 * sizeY / 4) - sizeFontY;
                break;
            case 5:
                possPosX = posX + (2 * sizeX / 4) - sizeFontX;
                possPosY = posY + (2 * sizeY / 4) - sizeFontY;
                break;
            case 6:
                possPosX = posX + (3 * sizeX / 4) - sizeFontX;
                possPosY = posY + (2 * sizeY / 4) - sizeFontY;
                break;

            case 7:
                possPosX = posX + (sizeX / 4) - sizeFontX;
                possPosY = posY + (3 * sizeY / 4) - sizeFontY;
                break;
            case 8:
                possPosX = posX + (2 * sizeX / 4) - sizeFontX;
                possPosY = posY + (3 * sizeY / 4) - sizeFontY;
                break;
            case 9:
                possPosX = posX + (3 * sizeX / 4) - sizeFontX;
                possPosY = posY + (3 * sizeY / 4) - sizeFontY;
                break;
            }

            wxString str;
            str.Printf(std2wx("%d"), possibility);

            dc->SetFont (*(board->getSmallFont()));
            //dc->SetTextForeground(this->fontColor);
            dc->SetTextForeground(*wxBLACK);

            dc->DrawText(str, possPosX, possPosY);
        }
    }

    dc->SetPen(*wxBLACK_PEN);
    dc->SetTextForeground(*wxBLACK);
}

bool Cell::isCell (wxPoint position)
{
    if (position.x > posX && position.x < (posX + sizeX))
    {
        if (position.y > posY && position.y < (posY + sizeY))
        {
            return true;
        }
    }
    return false;
}

void Cell::setHighlight (bool on)
{
    if (on)
    {
        color = *wxCYAN_BRUSH;
    }
    else
    {
        color = *wxWHITE_BRUSH;
    }
}

void Cell::highlight (Solver* solver)
{
    setHighlight(true);
    solver->doRefresh();
    solver->sleep();

    setHighlight(false);
    solver->doRefresh();
}

void Cell::highlightElements (Solver* solver)
{
    row->highlight (solver);
    column->highlight (solver);
    square->highlight (solver);
}

