// ============================================================================
//
// Copyright (c) 2010-2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ============================================================================


#ifndef CELL_H_
#define CELL_H_


#include <list>
#include <vector>

#include "Common.h"
#include "Square.h"
#include "Column.h"
#include "Row.h"

#ifndef BOARD_CLASS_
class Board;
#endif

#ifndef ROW_CLASS_
class Row;
#endif

#ifndef COLUMN_CLASS_
class Column;
#endif

#ifndef SQUARE_CLASS_
class Square;
#endif

#ifndef SOLVER_CLASS_
class Solver;
#endif


#ifndef CELL_CLASS_
#define CELL_CLASS_


class Cell
{
private:

    int value;
    int guess;

    bool hasPossibilities;
    bool hasPreviousGuesses;
    std::vector<int> possibilities;
    std::vector<int> previousGuesses;

    int colIdx;
    int rowIdx;
    Row* row;
    Column* column;
    Square* square;

    wxCoord posX;
    wxCoord posY;
    wxCoord sizeX;
    wxCoord sizeY;

    wxBrush color;
    CellType cellType;
	
    void bookValue (int value);
    void freeValue (int value);

public:
    Cell(int rowNumber, int colNumber);
    virtual ~Cell();

    const char* getName();

    void setRow (Row* row);
    void setColumn (Column* column);
    void setSquare (Square* square);
    bool setValue (int value, CellType cellType = sdST_INPUT, bool failSilently = false);
    int getValue();
    int getGuess();
    void reset();
    CellType getCellType ();

    int getRowIdx();
    int getColIdx();

    Row* getRow ();
    Column* getColumn ();
    Square* getSquare ();
    void initPosition (wxCoord posX, wxCoord posY, wxCoord sizeX, wxCoord sizeY);
    void draw (Board* board, wxDC* dc);
    bool isCell (wxPoint position);

    void highlight (Solver* solver);
    void setHighlight (bool on);
    void highlightElements (Solver* solver);

    bool isPossible (int value);
    bool containsValue (int value);
    bool isSinglePossibility(int value, int size);
    bool handleCellSinglePossibility(Solver *solver);
    bool hasValue();
    bool hasGuess();

    bool setPossibilities (int size);
    void addPossibility (int value);
    bool hasPossibility (int possibility);
    std::vector<int>* getPossibilities();
    bool discardPossibility(int value);
    void resetPossibilities();

    int makeGuess();
    bool allPossibilitiesTried();
    void fixGuess();

};

// a little inlining for the few little things called a lot and

inline int Cell::getRowIdx()
{
    return rowIdx;
}

inline int Cell::getColIdx()
{
    return colIdx;
}

inline int Cell::getValue()
{
    return value;
}

inline int Cell::getGuess()
{
    return guess;
}

inline bool Cell::hasValue()
{
    return this->value != 0;
}

inline bool Cell::hasGuess()
{
    return this->guess != 0;
}

inline bool Cell::containsValue (int value)
{
    return this->value == value || this->guess == value;
}

inline std::vector<int>* Cell::getPossibilities ()
{
    return &possibilities;
}

inline void Cell::bookValue (int value)
{
    this->row->bookValue(value);
    this->column->bookValue(value);
    this->square->bookValue(value);
}

inline void Cell::freeValue (int value)
{
    this->row->freeValue(value);
    this->column->freeValue(value);
    this->square->freeValue(value);
}

inline bool Cell::isPossible (int value)
{
    if (containsValue (value))
    {
        return true;
    }
    if (this->value != 0)
    {
        return false;
    }

    if (row->containsValue (value))
    {
        return false;
    }
    if (column->containsValue (value))
    {
        return false;
    }
    if (square->containsValue (value))
    {
        return false;
    }
    return true;
}

inline bool Cell::setValue (int newValue, CellType cellType, bool failSilently)
{
    if (newValue == 0)
    {
        if (this->value != 0)
        {
            this->freeValue(this->value);
        }
        if (this->guess != 0)
        {
            this->freeValue(this->guess);
        }

        this->value = 0;
        this->guess = 0;

        // reset possibilities
        if (hasPossibilities)
        {
            possibilities.clear();
            hasPossibilities = false;
        }
        if (hasPreviousGuesses)
        {
            previousGuesses.clear();
            hasPreviousGuesses = false;
        }
    }
    else
    {
        if (!isPossible(newValue))
        {
            if (failSilently) {
                return false;
            } else {
#ifdef __WXDEBUG__
                wxLogMessage (std2wx("throwing SoldokuException : Cell::setValue: Value %d is not possible for this cell."), newValue);
#endif
                char* msg = new char[100];
                sprintf(msg, "Cell::setValue: Value %d is not possible for this cell (%d, %d)", newValue, rowIdx, colIdx);
                throw new SoldokuException (msg);
            }
        } else {

            this->cellType = cellType;

            if (this->value != 0)
            {
                this->freeValue(this->value);
            }
            this->value = newValue;

            if (this->guess != 0)
            {
                this->freeValue(this->guess);
            }
            this->guess = 0;

            this->bookValue(newValue);

            // reset possibilities
            if (hasPossibilities)
            {
                possibilities.clear();
                hasPossibilities = false;
            }
            if (hasPreviousGuesses)
            {
                previousGuesses.clear();
                hasPreviousGuesses = false;
            }
        }
    }

    return true;
}

#endif



#endif
