// ============================================================================
//
// Copyright (c) 2010-2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ============================================================================


#include "Generator.h"

#include "Grid.h"
#include "Board.h"
#include "Solver.h"
#include "Cell.h"
#include "MainFrame.h"


#include <time.h>


// Utility functions

int* getRandomValueMapping(int size)
{
    int* retArray = new int[size];
    bool* assignedArray = new bool[size];

    for (int i = 0; i < size; i++)
    {
        assignedArray[i] = false;
    }

    for (int i = 0; i < size; i++)
    {
        int mappedValue = 0;

        do
        {
            mappedValue = random (0, size - 1);
        } while (assignedArray[mappedValue]);

        assignedArray[mappedValue] = true;
        retArray[i] = mappedValue;
    }

    return retArray;
}


// Generator class

Generator::Generator(Board* board, int size)
{
    this->board = board;
    this->size = size;

    time_t seconds;
    time(&seconds);
    srand((unsigned int) seconds);

    double squareSize = sqrt (size);

    DEFAULT_GRID = new Grid (size);

    for (int rowBaseCount = 0; rowBaseCount < size; rowBaseCount++)
    {
        int rowBase = rowBaseCount / squareSize;                     // 0 0 0 1 1 1 2 2 2
        int rowEff = (rowBaseCount % (int)squareSize) * squareSize;  // 0 3 6 0 3 6 0 3 6

        int rowCount = rowBase + rowEff;                             // 0 3 6 1 4 7 2 5 8

        for (int colCount = 0; colCount < size; colCount++)
        {
            int cellValue = (colCount + rowBaseCount) % size + 1;
            Cell* cell = DEFAULT_GRID->getCell(rowCount, colCount);
#ifdef __WXDEBUG__
            wxLogMessage(std2wx("Generator::Generator - default grid setting value %d for cell %s"), cellValue, ambstowcs(cell->getName()));
#endif
            cell->setValue(cellValue);
        }
    }

}

void Generator::swapCells(Cell *sourceCell, Cell *destCell)
{
    int sourceValue = sourceCell->getValue();
    int destValue = destCell->getValue();
    sourceCell->setValue(0);
    destCell->setValue(0);
    sourceCell->setValue(destValue);
    destCell->setValue(sourceValue);
}

Grid* Generator::generateGrid (int difficulty)
{

    // 1. Create new random solved grid
    Grid* originalGrid = newRandomGrid();


    // 2. Create initial grid ...
    Grid* retGrid = removeValuesOld (originalGrid);

    this->board->getMainFrame()->update();

#ifdef __WXDEBUG__
    wxLogMessage(std2wx("Generator::generateGrid - grid initially filled %d "), retGrid->countSettedCells());
#endif

    if (difficulty > 0)
    {

        // 3. Try adding some cells if they can make us get rid of some more cells
        tryReaddingValues(retGrid, originalGrid);

    }


    // This is stupid
/*
    int setCellsCnt = retGrid->countSettedCells();
#ifdef __WXDEBUG__
    wxLogMessage(std2wx("Generator::generateGrid - grid temp filled %d "), setCellsCnt);
    if (retGrid->isEasilySolveable()) {
        wxLogMessage(std2wx("Generator::generateGrid - Grid is easily solvable."));
    } else {
        wxLogMessage(std2wx("Generator::generateGrid - Grid is NOT easily solvable !!!"));
    }
#endif

    this->board->getMainFrame()->update();

    // Try to remove values than can be guessed
    removeGuessableValues (retGrid, originalGrid);

    this->board->getMainFrame()->update();
*/

    delete originalGrid;



    int setCellsCnt = retGrid->countSettedCells();
#ifdef __WXDEBUG__
    wxLogMessage(std2wx("Generator::generateGrid - grid finally filled %d "), setCellsCnt);
#endif

    char statusStr[50];
    sprintf (statusStr, "Grid filled with %d cells (%s grid)",  setCellsCnt, retGrid->isEasilySolveable() ? "easy" : "HARD");
    this->board->getMainFrame()->setStatusText(std2wx (statusStr));

#ifdef __WXDEBUG__
    if (retGrid->isEasilySolveable()) {
        wxLogMessage(std2wx("Generator::generateGrid - Grid is easily solvable."));
    } else {
        wxLogMessage(std2wx("Generator::generateGrid - Grid is NOT easily solvable !!!"));
    }
#endif



    // make sure grid is solveable with backtracking
    Solver* solver = new Solver(true, this->board);

    std::list<Grid*>* solutions = solver->findSolutions(retGrid);

    if (solutions->size() != 1)
    {
        char msg[50];
        sprintf (msg, "Internal problem : grid is not solvable (%d solutions)", (int) solutions->size());
        wxMessageDialog errorDlg(
                this->board, std2wx (msg), std2wx ("Error"));
        errorDlg.Center();
        errorDlg.ShowModal();
    }

    while (solutions->size() > 0)
    {
        delete solutions->front();
        solutions->pop_front();
    }
    delete solutions;


    // Done
    this->board->getMainFrame()->stopWait();

    return retGrid;
}

Grid* Generator::newRandomGrid ()
{
    double squareSize = sqrt (size);

    // 1. Build new grid from default one
    Grid* grid = new Grid (DEFAULT_GRID, false);// true); FIXME Doesn't work

    // 3. Random row permutations
    for (int squareCount = 0; squareCount < (int)squareSize; squareCount++)
    {
        // 2.0 A random number of permutations
        int numberOfPermutations = random (0, squareSize - 1);

        for (int i = 0; i < numberOfPermutations; i++)
        {
            // 2.1 Pick up one random row in square (between [squareCount * squareSize] and [(squareCount + 1) * squareSize - 1] )
            int firstRow = random (squareCount * squareSize, (squareCount + 1) * squareSize - 1);
            int secondRow = firstRow;
            while (secondRow == firstRow)
            {
                secondRow = random (squareCount * squareSize, (squareCount + 1) * squareSize - 1);
            }

            grid->switchRows (firstRow, secondRow);
        }
    }

    this->board->getMainFrame()->update();

    // 3. Random column permutations
    for (int squareCount = 0; squareCount < (int)squareSize; squareCount++)
    {
        // 3.0 A random number of permutations
        int numberOfPermutations = random (0, squareSize - 1);

        for (int i = 0; i < numberOfPermutations; i++)
        {
            // 3.1 Pick up one random column in square (between [squareCount * squareSize] and [(squareCount + 1) * squareSize - 1] )
            int firstColumn = random (squareCount * squareSize, (squareCount + 1) * squareSize - 1);
            int secondColumn = firstColumn;
            while (secondColumn == firstColumn)
            {
                secondColumn = random (squareCount * squareSize, (squareCount + 1) * squareSize - 1);
            }

            grid->switchColumns (firstColumn, secondColumn);
        }
    }

    this->board->getMainFrame()->update();

    // 4. Re-assign values

    // 4.1 Build a value mapping
    int* mapping = getRandomValueMapping(size);

    // 4.2 Build final Grid with Value Mapping
    Grid* tempGrid = new Grid (size);

    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            tempGrid->getCell(rowCounter, colCounter)->setValue(
                    mapping[grid->getCell(rowCounter, colCounter)->getValue() - 1] + 1);
        }
    }

    this->board->getMainFrame()->update();

    delete mapping;
    delete grid;

    return tempGrid;
}

void Generator::removeGuessableValues(Grid *retGrid, Grid *originalGrid)
{
    // now try to remove more cells. If grid remains solvable with 1 single guess, keep it away
    std::list<Cell*>* cellList = retGrid->listCellsWithValues();
    Solver* solver = new Solver(true, this->board);

    for( std::list<Cell*>::iterator it = cellList->begin(); // not listMyClass.begin()
        it != cellList->end(); // not listMyClass.end()
        it ++)
    {
        Cell* cell = (*it);
        int savedValue = cell->getValue();

        cell->setValue(0);

        Grid* copy = new Grid (retGrid);
        std::list<Grid*>* sols = solver->findSolutions(copy);
        if (sols->size() == 1) // 1 single guess is allowed
        {
            while (sols->size() > 0)
            {
                delete sols->front();
                sols->pop_front();
            }
            delete sols;

            break;
        }

        while (sols->size() > 0)
        {
            delete sols->front();
            sols->pop_front();
        }
        delete sols;

        delete copy;

        cell->setValue (savedValue);
    }

    delete cellList;
    delete solver;
}

void Generator::tryReaddingValues (Grid* retGrid, Grid* originalGrid)
{
    // At this stage I can try to add a value to see whether it helps me remove more
    bool betterFound = true;
    while (betterFound)
    {
        betterFound = false;

        std::list<Cell*>* cellListLev1 = retGrid->listCellsWithoutValues();

        while (cellListLev1->size() > 0)
        {
            Cell* cellLev1 = cellListLev1->front();
            cellListLev1->pop_front();

            cellLev1->setValue(originalGrid->getCell(cellLev1->getRowIdx(), cellLev1->getColIdx())->getValue());

            int newRemoveCount = performPossibleRemoves (retGrid, cellLev1);
            if (newRemoveCount > 0)
            {
                this->board->getMainFrame()->update();

                if (newRemoveCount > 1)
                {
#ifdef __WXDEBUG__
                    wxLogMessage(std2wx("Generator::tryReaddingValues - Found better cell (%d, %d) "), cellLev1->getRowIdx(), cellLev1->getColIdx());
#endif
                    betterFound = true;
                    break;
                }
            }
            else
            {
                cellLev1->setValue(0);
            }

            this->board->getMainFrame()->update();
        }

        delete cellListLev1;
    }

    this->board->getMainFrame()->update();

}

void Generator::addValuesRandomly (Grid* retGrid, Grid* originalGrid)
{
    int rowIndex = -1;
    int colIndex = -1;

    // 1. Add 10 values
    for (int i = 0; i < 3; i++)
    {
        rowIndex = random (0, size - 1);
        colIndex = random (0, size - 1);

        retGrid->getCell(rowIndex, colIndex)->setValue (originalGrid->getCell(rowIndex, colIndex)->getValue ());
    }
}

void Generator::addBestValuesUntilSolvable (Grid* retGrid, Grid* originalGrid)
{
    // 2. Add as much values as required to solve the grid
   while (!retGrid->isEasilySolveable())
   {
       // 2. Find cell with most possibilities
       Grid* tempGrid = new Grid (retGrid);
       Solver* solver = new Solver (true, NULL);
       tempGrid->solvePossibleCells (solver);
       Cell* cell = tempGrid->getRemainingCellWithMorePossibilities();

#ifdef __WXDEBUG__
       std::vector<int>* candidatePossibilities = cell->getPossibilities();
#endif
       int rowIndex = cell->getRow()->getNumber();
       int colIndex = cell->getColumn()->getNumber();

       // 3. Set cell value
#ifdef __WXDEBUG__
       wxLogMessage(std2wx("Generator::addBestValuesUntilSolvable (%d, %d - %d) "), rowIndex, colIndex, (int)candidatePossibilities->size());
#endif
       retGrid->getCell(rowIndex, colIndex)->setValue (originalGrid->getCell(rowIndex, colIndex)->getValue ());

       delete tempGrid;
       delete solver;
   }
}

void Generator::addOneCellForEachValue(Grid *& originalGrid, Grid *& retGrid)
{
    // 1. Add one value for each number
    for(int i = 0;i < this->size;i++)
    {
        std::vector<Cell*> *cellsWithValue = originalGrid->listCellsWithValue(i + 1);
        int idxToUse = random(0, cellsWithValue->size() - 1);
        Cell *origCell = cellsWithValue->at(idxToUse);
        retGrid->getCell(origCell->getRowIdx(), origCell->getColIdx())->setValue(origCell->getValue());
        delete cellsWithValue;
    }
}

Grid* Generator::removeValues (Grid* originalGrid)
{
    Grid* retGrid = new Grid (originalGrid);

    // 1. remove 20 cells randomly
    for (int i = 0; i < 20; i++)
    {
        int rowIdx = random (0, size - 1);
        int colIdx = random (0, size - 1);
        retGrid->getCell(rowIdx, colIdx)->setValue(0);
    }

    // 2. Not remove some more until grid becomes unsolvable
    //    following a specific approach
    Cell* cell = NULL;
    int possCountSearch = 0;
    int rowIdx = 0;
    int colIdx = 0;

    int startRowIdx = 0;
    int startColIdx = 0;

    while (retGrid->isSolveableBacktrack())
    {
        // 2.a If a made a complete loop without finding a new one of current
        //        possCountSearch, increment possCountSearch and restart
        if (rowIdx == startRowIdx && colIdx == startColIdx)
        {
            possCountSearch++;
        }

        // 2.b We are looking for a cell
        Cell* current = retGrid->getCell(rowIdx, colIdx);
        if (current->hasValue())
        {
            int oldValue = current->getValue();

            current->setValue(0);
            current->setPossibilities(size);

            std::vector<int>* possibilities = current->getPossibilities();
            if ( ((int)possibilities->size()) == possCountSearch)
            {
                startRowIdx = rowIdx;
                startColIdx = colIdx;
                cell = current;
            }
            else
            {
                current->setValue(oldValue);
            }

        }

        // 2.c Increment rowIdx and colIdx, handle looping case
        colIdx++;
        if (colIdx == 9)
        {
            colIdx = 0;
            rowIdx++;
            if (rowIdx == 9)
            {
                rowIdx = 0;
            }
        }
    }

    // 3. Re-add last one removed in order to make it solvable again
    cell->setValue(originalGrid->getCell(cell->getRowIdx(), cell->getColIdx())->getValue());

    // 4. Remove all than can be removed
    performPossibleRemoves (retGrid);

    return retGrid;
}

Grid* Generator::removeValuesOld (Grid* originalGrid)
{
    Grid* retGrid = new Grid (originalGrid);

    // 1. remove them all
    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            retGrid->getCell(rowCount, colCount)->setValue(0);
        }
    }

    // 1. Add one value for each number
    addOneCellForEachValue(originalGrid, retGrid);

    // 2. Add more back
    addBestValuesUntilSolvable (retGrid, originalGrid);

    //if (true) return retGrid;


    // 3. Try to find some more values to remove
    //int nbValuesRemoved =
    performPossibleRemoves (retGrid);

    return retGrid;
}


int Generator::performPossibleRemoves (Grid* retGrid, Cell* exception)
{
    int retValue = 0;

    std::list<Cell*>* cellList = retGrid->listCellsWithValues();

    // randomly test all remaining values whether they can be removed
    for( std::list<Cell*>::iterator it = cellList->begin(); // not listMyClass.begin()
        it != cellList->end(); // not listMyClass.end()
        it ++)
    {
        Cell* cell = (*it);

        if (exception == NULL || cell != exception)
        {
            int savedValue = cell->getValue();
            cell->setValue(0);


            /* This is an attempt to speed up the process by making it not
             * trying cells that have only low probability of being able to
             * be removed.
             */
            bool tryCell = true;
            if (exception != NULL)
            {
                cell->setPossibilities(size);
                std::vector<int>* cellPoss = cell->getPossibilities();
                int possSize = cellPoss->size();
                cell->resetPossibilities();
                tryCell = possSize <= 5;
            }


            if (!tryCell || !retGrid->isSolveableBacktrack())
            {
                cell->setValue(savedValue);
            }
            else
            {
                retValue++;
#ifdef __WXDEBUG__
                wxLogMessage(std2wx("Generator::performPossibleRemoves - Removing cell (%d, %d) "), cell->getRowIdx(), cell->getColIdx());
#endif
            }
        }
    }

    delete cellList;

    return retValue;
}


