// ============================================================================
//
// Copyright (c) 2010-2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ============================================================================



#include "Grid.h"

#include "Common.h"
#include "Solver.h"
#include "SolverThread.h"
#include "Row.h"
#include "Column.h"
#include "Square.h"
#include "Group.h"
#include "InitValuesHolder.h"



Grid::Grid(int size)
{
    init(size);
}

Grid::Grid(int size, int** initialValues)
{
    init(size);

    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            Cell* cell = getCell(rowCounter, colCounter);
            cell->setValue(initialValues[rowCounter][colCounter]);
        }
    }
}

Grid::Grid (Grid* original, bool randomize)
{
    init(original->getSize());

    int base = 0;
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        if (randomize)
        {
            base += random (0, 1);
        }
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            Cell* cell = original->getCell(rowCounter, colCounter);
            Cell* target = getCell(rowCounter, colCounter);
            if (cell->hasValue())
            {
                target->setValue(
                        randomize ? ((cell->getValue() + base) % 9) + 1 : cell->getValue(),
                        cell->getCellType());
            }
            else if (cell->hasGuess())
            {
                target->setValue(
                        randomize ? ((cell->getGuess() + base) % 9) + 1 : cell->getGuess(),
                        sdST_GUESS);
            }
            else
            {
                std::vector<int>* cellPoss = cell->getPossibilities();
                for( std::vector<int>::iterator it = cellPoss->begin(); // not listMyClass.begin()
                    it != cellPoss->end(); // not listMyClass.end()
                    ++it)
                {
                    int i = (*it);
                    target->addPossibility(i);
                }
            }
        }
    }
}

Grid::~Grid()
{
    destroy();
}

void Grid::destroy()
{
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            delete cells[rowCounter][colCounter];
        }

        delete cells[rowCounter];
        delete rows[rowCounter];
    }

    delete cells;
    delete rows;

    // 2. Build columns
    for (int colCounter = 0; colCounter < size; colCounter++)
    {
        delete columns[colCounter];
    }
    delete columns;

    // 3. Build squares
    double squareSize = sqrt (size);
    int squareCounter = 0;
    for (int rowCounter = 0; rowCounter < size; rowCounter += (int)squareSize)
    {
        for (int colCounter = 0; colCounter < size; colCounter += (int)squareSize)
        {
            delete squares[squareCounter];
            squareCounter++;
        }
    }
}

void Grid::init(int size)
{
    started = false;

    double squareSize = sqrt (size);

    // make sure the sqrt is an entire number
    if (floor (squareSize) != squareSize)
    {
#ifdef __WXDEBUG__
        wxLogMessage (std2wx("throwing SoldokuException : Grid::Grid : provided size is not an entire number"));
#endif
        throw new SoldokuException ((char*)"Grid::Grid : provided size is not an entire number");
    }

    this->size = size;

    cells = new Cell**[size];

    rows = new Row*[size];
    columns = new Column*[size];
    squares = new Square*[size];

    // 1. Build rows and cells
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        cells[rowCounter] = new Cell*[size];
        rows[rowCounter] = new Row(size, rowCounter);

        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            cells[rowCounter][colCounter] = new Cell (rowCounter, colCounter);
        }

        rows[rowCounter]->init(this);
    }

    // 2. Build columns
    for (int colCounter = 0; colCounter < size; colCounter++)
    {
        columns[colCounter] = new Column(size, colCounter);
        columns[colCounter]->init(this);
    }

    // 3. Build squares
    int squareCounter = 0;
    for (int rowCounter = 0; rowCounter < size; rowCounter += (int)squareSize)
    {
        for (int colCounter = 0; colCounter < size; colCounter += (int)squareSize)
        {
            squares[squareCounter] = new Square ((int)squareSize, rowCounter, colCounter);
            squares[squareCounter]->init(this);
            squareCounter++;
        }
    }
}

void Grid::draw (Board* board, wxDC* dc, wxSize size)
{
    // don't draw anything of we don't have the place for it
    if (size.GetX() > 500 && size.GetY() > 500)
    {
        // we take 40 pixels border
        int gridSizeW = size.GetX() - 60;
        int gridSizeH = size.GetY() - 60;

        // draw the body
        dc->SetPen(*wxGREY_PEN);
        dc->SetBrush(*wxGREEN_BRUSH);

        // draw border
        int cellSizeW = gridSizeW / this->size;
        int cellSizeH = gridSizeH / this->size;

        // draw cells
        for (int rowCounter = 0; rowCounter < this->size; rowCounter++)
        {
            int rowOffset = 30 + rowCounter * cellSizeH;

            for (int colCounter = 0; colCounter < this->size; colCounter++)
            {
                int colOffset = 30 + colCounter * cellSizeW;

                this->cells[rowCounter][colCounter]->initPosition (colOffset, rowOffset, cellSizeW, cellSizeH);

                this->cells[rowCounter][colCounter]->draw (board, dc);
            }
        }

        // draw squares
        double squareNumber = sqrt (this->size);

        int squareOffsetW = cellSizeW * squareNumber;
        int squareOffsetH = cellSizeH * squareNumber;

        int squareSizeW = (cellSizeW * squareNumber) - 2;
        int squareSizeH = (cellSizeH * squareNumber) - 2;

        int squareCounter = 0;
        for (int rowCounter = 0; rowCounter < squareNumber; rowCounter++)
        {
            int rowOffset = 31 + rowCounter * squareOffsetH;

            for (int colCounter = 0; colCounter < squareNumber; colCounter++)
            {
                int colOffset = 31 + colCounter * squareOffsetW;

                squares[squareCounter]->initPosition (colOffset, rowOffset, squareSizeW, squareSizeH);
                squares[squareCounter]->draw (dc);
                squareCounter++;
            }
        }

    } 
    else 
    {
       // FIXME We should do something in this case
    }

}

int Grid::countSettedCells()
{
    int retValue = 0;

    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            if (cells[rowCounter][colCounter]->getValue() != 0)
            {
                retValue++;
            }
        }
    }

    return retValue;
}

Cell* Grid::getCell (wxPoint position)
{
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            if (cells[rowCounter][colCounter]->isCell (position))
            {
                return cells[rowCounter][colCounter];
            }
        }
    }
    return NULL;
}

void Grid::switchRows (int first, int second)
{
    // 1. Make sure both rows belong to grid
    if (first < 0 || first >= size)
    {
#ifdef __WXDEBUG__
        wxLogMessage (std2wx("throwing SoldokuException : First row is outside grid range"));
#endif
        throw new SoldokuException ((char*)"First row is outside grid range");
    }
    if (second < 0 || second >= size)
    {
#ifdef __WXDEBUG__
        wxLogMessage (std2wx("throwing SoldokuException : Second row is outside grid range"));
#endif
        throw new SoldokuException ((char*)"Second row is outside grid range");
    }

    // 2. Make sure both rows belong to same horizontal squares
    double squareSize = sqrt (size);
    int firstSquare = first / (int) squareSize;
    int secondSquare = second / (int) squareSize;
    if (firstSquare != secondSquare)
    {
#ifdef __WXDEBUG__
        wxLogMessage (std2wx("throwing SoldokuException : First and second rows not within same squares group"));
#endif
        throw new SoldokuException ((char*)"First and second rows not within same squares group");
    }

    // 3. Actually switch them
    std::vector<int> firstBackup;
    std::vector<int> secondBackup;
    for (int i = 0; i < size; i++)
    {
        firstBackup.push_back(cells[first][i]->getValue());
        secondBackup.push_back(cells[second][i]->getValue());

        cells[first][i]->setValue(0);
        cells[second][i]->setValue(0);
    }

    for (int i = 0; i < size; i++)
    {
        cells[first][i]->setValue(secondBackup[i]);
        cells[second][i]->setValue(firstBackup[i]);
    }
}

void Grid::switchColumns (int first, int second)
{
    // 1. Make sure both columns belong to grid
    if (first < 0 || first >= size)
    {
#ifdef __WXDEBUG__
        wxLogMessage (std2wx("throwing SoldokuException : First column is outside grid range"));
#endif
        throw new SoldokuException ((char*)"First column is outside grid range");
    }
    if (second < 0 || second >= size)
    {
#ifdef __WXDEBUG__
        wxLogMessage (std2wx("throwing SoldokuException : Second column is outside grid range"));
#endif
        throw new SoldokuException ((char*)"Second column is outside grid range");
    }

    // 2. Make sure both rows belong to same horizontal squares
    double squareSize = sqrt (size);
    int firstSquare = first / (int) squareSize;
    int secondSquare = second / (int) squareSize;
    if (firstSquare != secondSquare)
    {
#ifdef __WXDEBUG__
        wxLogMessage (std2wx("throwing SoldokuException : First and second columns not within same squares group"));
#endif
        throw new SoldokuException ((char*)"First and second columns not within same squares group");
    }

    // 3. Actually switch them
    std::vector<int> firstBackup;
    std::vector<int> secondBackup;
    for (int i = 0; i < size; i++)
    {
        firstBackup.push_back(cells[i][first]->getValue());
        secondBackup.push_back(cells[i][second]->getValue());

        cells[i][first]->setValue(0);
        cells[i][second]->setValue(0);
    }

    for (int i = 0; i < size; i++)
    {
        cells[i][first]->setValue(secondBackup[i]);
        cells[i][second]->setValue(firstBackup[i]);
    }
}

bool Grid::isSolveableBacktrack()
{
    // If grid has fewer than 17 values set, no need to try it any further
    std::list<Cell*>* cellList = listCellsWithValues();
    if (cellList->size() < 17)
    {
        delete cellList;
        return false;
    }
    delete cellList;

    Solver* solver = new Solver(true, NULL);

    bool retValue = solver->isSolvable(this);

    delete solver;

    return retValue;
}

bool Grid::isEasilySolveable()
{
    Grid* copy = new Grid (this);

    Solver* solver = new Solver(true, NULL);
    copy->solvePossibleCells(solver);

    bool retValue = copy->isSolutionFinished();

    delete copy;
    delete solver;

    return retValue;
}

void Grid::start(Solver* solver)
{
    try
    {
        solve(solver);
    }
    catch (SoldokuException* e)
    {
        char message[1000];
        strcat (message, "An error occured during solving. \nBoard will be reset. \nError was : \n  ");
        strcat (message, e->what());
        solver->handleSolveError((const char *) &message);

        delete e;
    }

    solver->doRefresh();
}

void Grid::solve(Solver* solver)
{
    // Phase A : Constructive algorithm
    solvePossibleCells (solver);

    // Phase B :
    solveUsingGuess (solver);

    solvePossibleCells (solver);
}

bool Grid::isSolutionFinished()
{
    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            if (!cells[rowCount][colCount]->hasValue())
            {
                return false;
            }
        }
    }

    return true;
}

std::vector<Cell*>* Grid::listCellsWithValue(int value)
{
    std::vector<Cell*>* retList = new std::vector<Cell*>();

    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            Cell* cell = cells[rowCount][colCount];
            if (cell->getValue() == value)
            {
                retList->push_back (cell);
            }
        }
    }

    return retList;
}

std::list<Cell*>* Grid::listCellsWithoutValues()
{
    std::list<Cell*>* retList = new std::list<Cell*>();

    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            Cell* cell = cells[rowCount][colCount];
            if (!cell->hasValue())
            {
                // shuffle a little
                if (colCount % 2 == 0)
                {
                    retList->push_back (cell);
                }
                else
                {
                    retList->push_front(cell);
                }
            }
        }
    }

    return retList;
}

std::list<Cell*>* Grid::listCellsWithValues()
{
    std::list<Cell*>* retList = new std::list<Cell*>();

    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            Cell* cell = cells[rowCount][colCount];
            if (cell->hasValue())
            {
                // shuffle a little
                if (colCount % 2 == 0)
                {
                    retList->push_back (cell);
                }
                else
                {
                    retList->push_front(cell);
                }
            }
        }
    }

    return retList;
}

Cell* Grid::getRemainingCellWithLessPossibilities ()
{
    Cell* retCell = NULL;
    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            Cell* cell = cells[rowCount][colCount];

            if (!cell->hasValue() && !cell->hasGuess() && cell->getPossibilities()->size() > 0)
            {
                if (retCell == NULL)
                {
                    retCell = cell;
                }
                else if (!cell->hasGuess() && !cell->hasValue())
                {
                    std::vector<int>* candidatePossibilities = cell->getPossibilities();
                    std::vector<int>* retCellPossibilities = retCell->getPossibilities();

                    if (candidatePossibilities->size() < retCellPossibilities->size())
                    {
                        retCell = cell;
                    }
                }
            }
        }
    }

    return retCell;
}

Cell* Grid::getRemainingCellWithMorePossibilities ()
{
    Cell* retCell = NULL;
    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            Cell* cell = cells[rowCount][colCount];

            if (cell->getPossibilities()->size() > 0)
            {
                if (retCell == NULL)
                {
                    retCell = cell;
                }
                else if (!cell->hasGuess() && !cell->hasValue())
                {
                    std::vector<int>* candidatePossibilities = cell->getPossibilities();
                    std::vector<int>* retCellPossibilities = retCell->getPossibilities();

                    if (candidatePossibilities->size() > retCellPossibilities->size())
                    {
                        retCell = cell;
                    }
                }
            }
        }
    }

    return retCell;
}

void Grid::solveUsingGuess (Solver* solver, int maxRecursiveCalls)
{
#ifdef __WXDEBUG__
	std::cout << "solveUsingGuess ..." << std::endl;
#endif

    if (solver->shouldCont())
    {
        // 1. Take the cell with the littlest number of possibilities
        Cell* cell = getRemainingCellWithLessPossibilities();

        // 2. while not all possibilities found and no exception occurred
        bool lastOneFailed = false;
        while (cell != NULL && !cell->allPossibilitiesTried())
        {
            lastOneFailed = false;
            if (solver->shouldCont())
            {
                try
                {
                    // 3.a Make a guess for this cell
                    int guess = cell->makeGuess();
                    if (guess != 0)
                    {
                        // 3.b Solve the usual way
                        Grid* gridCopy = new Grid(this);
                        if (!solver->isFastResolution())
                        {
                            solver->displayGrid (gridCopy);
                        }

                        try
                        {
                            gridCopy->solvePossibleCells (solver);
                        }

                        // 3.c.1 if an exceptions is thrown, we can't solve using this guess,
                        // have to try another one
                        catch (SoldokuException* e)
                        {
                            delete e;
                            if (!solver->isFastResolution())
                            {
                                solver->displayGrid (this);
                            }
                            delete gridCopy;
                            lastOneFailed = true;
                            continue;
                        }

                        // 3.c.2 If finished
                        if (gridCopy->isSolutionFinished())
                        {
                            // 3.c.2.a set guess for real
                            cell->fixGuess();

                            // 3.c.2.b return
                            if (!solver->isFastResolution())
                            {
                                solver->displayGrid (this);
                            }
                            delete gridCopy;
                            break;

                        }
                        // 3.c.3 solution not finished
                        else
                        {
                            // 3.c.3.a call recursively
                            solveUsingGuess (solver, maxRecursiveCalls - 1);

                            // 3.c.3.a call if no exception occurred, it could be solved using
                            //         this call, let's fix the value
                            cell->fixGuess();
                        }

                        if (!solver->isFastResolution())
                        {
                            solver->displayGrid (this);
                        }
                        delete gridCopy;
                    }
                }
                catch (SoldokuException* e)
                {
#ifdef __WXDEBUG__
                    std::cout << e->what() << std::endl;
#endif
                    lastOneFailed = false;
                    delete e;
                }
            }
        }
        if (lastOneFailed)
        {
            throw new SoldokuException ((char*)"Unsolvable.");
        }
    }
}

void Grid::solvePossibleCells (Solver* solver)
{
#ifdef __WXDEBUG__
	std::cout << "solvePossibleCells ..." << std::endl;
#endif

    bool changesMade;
    do
    {
        changesMade = false;

        // 0. reset all possibilities
        resetPossibilities (solver);

        // 1. Solve all single possibilities
        if (solveSinglePossibilities(solver))
        {
            changesMade = true;
            while (solveSinglePossibilities(solver));
        }

        // This is now a second stage of the algorithm where we work with
        // possibilities
        // Here we track at the end the changes we have been able to made thanks
        // to possibilities discarding approached below.
        // If not changes have been made, it means we have reached a dead end
        // with the currently implemented techniques

        // 2. Set 4 possibilities and less
        setPossiblities(solver);

        bool possibilityChangesMade;
        do
        {
            possibilityChangesMade = false;

            // 3. Solve possibilities based on groups
            if (solvePossiblities(solver))
            {
                possibilityChangesMade = true;
            }

            // 4. Solve possibilities on rows/columns that discards possibilities
            //    on squares
            if (solveSquarePossiblities(solver))
            {
                possibilityChangesMade = true;
            }

        } while (possibilityChangesMade);

        // X. Now set remaining single possibilities
        if (setRemainingPossibilities(solver)) changesMade = true;

    } while (changesMade);
}

void Grid::resetPossibilities(Solver* solver)
{
    // Loop through each cell
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            if (solver->shouldCont())
            {
                cells[rowCounter][colCounter]->resetPossibilities();
            }
        }
    }

}

bool Grid::setRemainingPossibilities(Solver* solver)
{
    bool changesMade = false;

    // Loop through each cell
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            if (solver->shouldCont())
            {
                if (!cells[rowCounter][colCounter]->hasValue())
                {
                    std::vector<int>* possibilities = cells[rowCounter][colCounter]->getPossibilities();

                    if (possibilities->size() == 1)
                    {

                        if (!solver->isFastResolution())
                        {
                            cells[rowCounter][colCounter]->highlight(solver);


                            char statusStr[50];
                            sprintf (statusStr, "Setting remaining possibility on Cell %s",  cells[rowCounter][colCounter]->getName());
                            solver->setStatusText(std2wx(statusStr));
                        }

                        cells[rowCounter][colCounter]->setValue ((*possibilities)[0], sdST_POSS);
                        changesMade = true;
                    }
                }
            }
        }
    }

    return changesMade;
}

bool Grid::solveSquarePossiblities(Solver* solver)
{
    bool changesMade = false;
    if (isSolutionFinished())
    {
        return false;
    }

    for (int counter = 0; counter < size; counter++)
    {
        if (solveSquareByElementPossibilities (rows[counter], solver))
        {
            changesMade = true;
        }
    }

    for (int counter = 0; counter < size; counter++)
    {
        if (solveSquareByElementPossibilities (columns[counter], solver))
        {
            changesMade = true;
        }
    }

    return changesMade;
}

bool Grid::solveSquareByElementPossibilities(Element* element, Solver* solver)
{
    bool changesMade = false;
    if (solver->shouldCont())
    {
        if (!solver->isFastResolution())
        {
            char statusStr[50];
            sprintf (statusStr, "Solving Square possibilities with %s", element->getName());
            solver->setStatusText(std2wx(statusStr));

            element->highlight(solver);
        }

        for (int i = 0; i < size; i++)
        {
            std::vector<int>* possGroup = element->getCellPossibilities (i);

            // loop through each group (4 or less possibilities in a cell)

            int possGroupSize = possGroup->size();
            if (possGroupSize >= 2 && possGroupSize <= 4)
            {
                Group* group = element->getCellsWithSamePoss (possGroup);

                //wxLogMessage ("Grid::solveSquareByElementPossibilities group %s for cell %d has group size %d and %d same groups.",
                //        element->getName(), i , possGroupSize, group->size());

                // congratulations. We have a group
                if (group->size() >= possGroupSize)
                {

                    Square* square = element->getMatchingSquare (i);

                    if (!solver->isFastResolution())
                    {
                        group->highlight(solver);
                    }

                    // loop through each possibility in group (4 or less possibilities in a cell)
                    for (int possCounter = 0; possCounter < possGroupSize; possCounter++)
                    {
                        int possibility = (*possGroup)[possCounter];

                        bool allGood = true;

                        // search for other cells within group with same possibility
                        std::vector<int>* cellsWithSamePoss = group->listCellsWithPossibility (possibility);

                        // if they are all within same square
                        int otherCellsSize = cellsWithSamePoss->size();
                        if (otherCellsSize == 0)
                        {
                            allGood = false;
                        }
                        else
                        {
                            for (int otherCellsCounter = 0; otherCellsCounter < otherCellsSize; otherCellsCounter++)
                            {
                                Square* otherCellSquare = element->getMatchingSquare (cellsWithSamePoss->operator [](otherCellsCounter));
                                if (otherCellSquare != square)
                                {
                                    allGood = false;
                                }
                            }
                        }

                        // the possibility can be discarded from other square cells
                        if (allGood)
                        {
                            if (!solver->isFastResolution())
                            {
                                char statusStr[50];
                                sprintf (statusStr, "  Cell  %d discards possibility %d ( %d cells with same pos)", i, possibility, otherCellsSize);
                                solver->setStatusText(std2wx(statusStr));
                            }

                            if (square->discardPossibilityFromOther (possibility, element))
                            {
                                changesMade = true;
                            }
                        }

                        delete cellsWithSamePoss;
                    }
                }

                delete group;
            }
        }
    }

    return changesMade;
}

bool Grid::solvePossiblities(Solver* solver)
{
    bool changesMade = false;
    if (isSolutionFinished())
    {
        return false;
    }

    for (int i = 0; i < size; i++)
    {
        if (solveElementPossiblities (rows[i], solver))
        {
            changesMade = true;
        }
    }

    for (int i = 0; i < size; i++)
    {
        if (solveElementPossiblities (columns[i], solver))
        {
            changesMade = true;
        }
    }

    for (int i = 0; i < size; i++)
    {
        if (solveElementPossiblities (squares[i], solver))
        {
            changesMade = true;
        }
    }
    return changesMade;
}


bool Grid::solveElementPossiblities(Element* element, Solver* solver)
{
    bool changesMade = false;
    if (solver->shouldCont())
    {
        if (!solver->isFastResolution())
        {
            char statusStr[50];
            sprintf (statusStr, "Solving possibilities on %s", element->getName());
            solver->setStatusText(std2wx(statusStr));

            element->highlight (solver);
        }

        for (int i = 0; i < size; i++)
        {
            std::vector<int>* possGroup = element->getCellPossibilities (i);

            // loop through each group (4 or less possibilities in a cell)
            int possGroupSize = possGroup->size();
            if (possGroupSize >= 2 && possGroupSize <= 4)
            {
                Group* group = element->getCellsWithSamePoss (possGroup);

                //wxLogMessage ("Grid::solveElementPossiblities group %s for cell %d has group size %d and %d same groups.",
                //        element->getName(), i , possGroupSize, group->size());


                // if (group size) number of cells are found, discard group members from other cells
                if (group->size() >= possGroupSize)
                {
                    if (!solver->isFastResolution())
                    {
                        // highlight group element
                        group->highlight (solver);
                    }

                    for (int j = 0; j < size; j++)
                    {
                        if (j != i && !group->contains (j))
                        {
                            int groupSize = possGroup->size();
                            for (int valueCounter = 0; valueCounter < groupSize; valueCounter++)
                            {
                                if (element->discardCellPossibility (j, (*possGroup)[valueCounter]))
                                {
                                    changesMade = true;
                                }
                            }
                        }
                    }
                }

                delete group;
            }
        }
    }

    return changesMade;
}

void Grid::setPossiblities(Solver* solver)
{
    // Loop through each cell
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            if (solver->shouldCont())
            {
                if (!cells[rowCounter][colCounter]->hasValue())
                {
                    if (!solver->isFastResolution())
                    {
                        char statusStr[50];
                        sprintf (statusStr, "Setting possibilities on Cell (%d, %d)", rowCounter, colCounter);
                        solver->setStatusText(std2wx(statusStr));

                        cells[rowCounter][colCounter]->highlight(solver);
                    }

                    cells[rowCounter][colCounter]->setPossibilities (size);
                }
            }
        }
    }
}

bool Grid::solveSinglePossibilities(Solver* solver)
{
    bool retValue = false;

    // Loop through each cell
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            Cell* cell =  cells[rowCounter][colCounter];
            if (!cell->hasValue())
            {
                if (cell->handleCellSinglePossibility(solver))
                {
                    retValue = true;
                }
            }
        }
    }

    return retValue;
}

void Grid::clean ()
{
    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            cells[rowCount][colCount]->setValue(0);
        }
    }
}

void Grid::reset ()
{
    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            cells[rowCount][colCount]->reset();
        }
    }
}

void Grid::initialize (InitValuesHolder* holder)
{
    // 1. Reinit Grid
    clean();

    // 2. Set loaded values
    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            CellType cellType = holder->getCellType(rowCount, colCount);
            int cellValue = holder->getCellValue(rowCount, colCount);
            if (cellValue != 0)
            {
#ifdef __WXDEBUG__
                wxLogMessage (std2wx("Grid::initialize - Cell (%d, %d) : setting value : %d"), rowCount, colCount, cellValue);
#endif
                cells[rowCount][colCount]->setValue(cellValue, cellType);
            }

            std::vector<int>* possibilities = holder->getCellPossibilities(rowCount, colCount);
            if (possibilities != NULL && possibilities->size() > 0)
            {
                for( std::vector<int>::iterator it = possibilities->begin(); // not listMyClass.begin()
                    it != possibilities->end(); // not listMyClass.end()
                    ++it)
                {
                    int i = (*it);
                    cells[rowCount][colCount]->addPossibility(i);
                }
            }
        }
    }
}

InitValuesHolder* Grid::getGridValuesHolder ()
{
    InitValuesHolder* holder = new InitValuesHolder(size);

    for (int rowCount = 0; rowCount < size; rowCount++)
    {
        for (int colCount = 0; colCount < size; colCount++)
        {
            Cell* cell = cells[rowCount][colCount];
            holder->setCellValue(rowCount, colCount, cell->getValue());
            holder->setCellType(rowCount, colCount, cell->getCellType());

            std::vector<int>* poss = new std::vector<int>(*(cell->getPossibilities()));

            holder->setCellPossibilities(rowCount, colCount, poss);
        }
    }

    return holder;
}

void Grid::printOnConsole()
{
    wxLogMessage (std2wx("---------------"));
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        wxLogMessage (std2wx("|%d%d%d|%d%d%d|%d%d%d|"),
                cells[rowCounter][0]->getValue(),
                cells[rowCounter][1]->getValue(),
                cells[rowCounter][2]->getValue(),
                cells[rowCounter][3]->getValue(),
                cells[rowCounter][4]->getValue(),
                cells[rowCounter][5]->getValue(),
                cells[rowCounter][6]->getValue(),
                cells[rowCounter][7]->getValue(),
                cells[rowCounter][8]->getValue()
                );

        if (rowCounter > 0 && ((rowCounter + 1) % 3) == 0)
        {
            wxLogMessage (std2wx("---------------"));
        }
    }
}

void Grid::stop()
{
    for (int rowCounter = 0; rowCounter < size; rowCounter++)
    {
        for (int colCounter = 0; colCounter < size; colCounter++)
        {
            cells[rowCounter][colCounter]->setHighlight(false);
        }
    }
}

