// ============================================================================
//
// Copyright (c) 2010-2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ============================================================================


#include "Solver.h"

#include "Board.h"


Solver::Solver (bool fastResolution, Board* board)
{
    this->fastResolution = fastResolution;
    this->board = board;    
}

bool Solver::isFastResolution()
{
    return fastResolution;
}

Solver::~Solver()
{
    this->board = NULL;
}

void Solver::showError (const char* message)
{
    this->board->showError (message);
}

void Solver::handleSolveError (const char* message)
{
    // This is buggy
    this->board->showError (message);
    this->board->resetGrid();
}

bool Solver::shouldCont (void)
{
    return true;
}

void Solver::displayGrid (Grid* grid)
{
    if (!fastResolution)
    {
        board->displayGrid(grid);
    }
}

void Solver::doRefresh ()
{
    if (!fastResolution)
    {
        board->doRefresh();
    }
}

void Solver::sleep ()
{
    if (!fastResolution)
    {
        board->sleep();
    }
}

void Solver::setStatusText (const wxString& text)
{
    if (!fastResolution)
    {
        board->setStatusText(text);
    }    
}

bool Solver::isSolvable(Grid* source)
{

    Grid* original = new Grid (source);
    try
    {
        original->solvePossibleCells(this);
    }
    catch (SoldokuException* e)
    {
        delete e;
        return false;
    }

    Grid* copy = new Grid (original);

    std::list<Cell*>* emptyCells = copy->listCellsWithoutValues();
    int ecSize = emptyCells->size();
    Cell** emptyCellsArray = new Cell*[ecSize];
    for (int i = 0; i < ecSize; i++)
    {
        emptyCellsArray[i] = emptyCells->front();
        emptyCells->pop_front();
    }
    int count = 0;
    __isSolvable (&count, 0, ecSize, emptyCellsArray, copy, original);

    delete emptyCellsArray;
    delete emptyCells;
    delete copy;
    delete original;

    return count == 1;
}

void Solver::__isSolvable(int* count, int idx, int ecSize, Cell** emptyCellsArray, Grid* source, Grid* original)
{
    if (ecSize == 0 && source->isSolutionFinished())
    {
        (*count)++;
        return;
    }

    // find next a cell without a value
    if (idx >= ecSize)
    {
        return;
    }
    Cell* cell = emptyCellsArray[idx];

    if (cell != NULL)
    {
        // try all values for it
        Cell* origCell = original->getCell(cell->getRowIdx(), cell->getColIdx());
        std::vector<int>* cellPoss = origCell->getPossibilities();

        if (cellPoss->size() > 0)
        {
            for( std::vector<int>::iterator it = cellPoss->begin(); // not listMyClass.begin()
                it != cellPoss->end(); // not listMyClass.end()
                ++it)
            {
                int i = (*it);

                if (cell->setValue(i, sdST_INPUT, true))
                {
                    // if grid is solved, increment count of solutions
                    if (idx >= ecSize - 1 && source->isSolutionFinished())
                    {
                        (*count)++;
                    }
                    // otherwise call recursively
                    else
                    {
                        __isSolvable(count, idx + 1, ecSize, emptyCellsArray, source, original);
                    }

                    // reset it before returning
                    cell->setValue(0);

                    if ( (*count) > 1)
                    {
                        return;
                    }
                }
            }
        }
    }
}

std::list<Grid*>* Solver::findSolutions(Grid* source)
{
    Grid* original = new Grid (source);
    original->solvePossibleCells(this);

    Grid* copy = new Grid (original);

    std::list<Grid*>* retList = new std::list<Grid*>; // Make a copy before adding to list

    std::list<Cell*>* emptyCells = copy->listCellsWithoutValues();
    int ecSize = emptyCells->size();
    Cell** emptyCellsArray = new Cell*[ecSize];
    for (int i = 0; i < ecSize; i++)
    {
        emptyCellsArray[i] = emptyCells->front();
        emptyCells->pop_front();
    }
    __findSolutions (retList, 0, ecSize, emptyCellsArray, copy, original);

    delete emptyCellsArray;
    delete emptyCells;
    delete copy;

    return retList;
}

void Solver::__findSolutions(std::list<Grid*>* retList, int idx, int ecSize, Cell** emptyCellsArray, Grid* source, Grid* original)
{
    if (ecSize == 0 && source->isSolutionFinished())
    {
        Grid* copy = new Grid (source);
        retList->push_back(copy);
        return;
    }

    // find a cell without a value
    if (idx >= ecSize)
    {
        return;
    }
    Cell* cell = emptyCellsArray[idx];

    // try all values for it
    Cell* origCell = original->getCell(cell->getRowIdx(), cell->getColIdx());
    std::vector<int>* cellPoss = origCell->getPossibilities();

    if (cellPoss->size() > 0)
    {
        for( std::vector<int>::iterator it = cellPoss->begin(); // not listMyClass.begin()
            it != cellPoss->end(); // not listMyClass.end()
            ++it)
        {
            int i = (*it);

            if (cell->setValue(i, sdST_INPUT, true))
            {
                // if grid is solved, add a copy to retList
                if (idx >= ecSize - 1 && source->isSolutionFinished())
                {
                    Grid* copy = new Grid (source);
                    retList->push_back(copy);
                }
                // otherwise call recursively
                else
                {
                    __findSolutions(retList, idx + 1, ecSize, emptyCellsArray, source, original);
                }

                // reset it before returning
                cell->setValue(0);
            }
        }
    }
}
